
import brownian_diffuser
import neural_diffeqs
import torchdiffeq
import torchsde


class Credentials:
    def __init__(self, func, adjoint=False):
        self.func = func
        self._adjoint = adjoint

    @property
    def is_NeuralODE(self):
        return isinstance(self.func, neural_diffeqs.NeuralODE)

    @property
    def is_NeuralSDE(self):
        return isinstance(self.func, neural_diffeqs.NeuralSDE)

    @property
    def is_TorchNet(self):
        return (not self.is_NeuralODE) and (not self.is_NeuralSDE)

    @property
    def integrator(self):
        return self._integrator

    @property
    def mu_is_potential_func(self):
        """Assumes potential is 1-D"""
        if not self.is_TorchNet:
            return list(self.func.mu.parameters())[-1].shape[0] == 1
        return list(self.func.parameters())[-1].shape[0] == 1
    
    @property
    def sigma_is_potential_func(self):
        """Assumes potential is 1-D"""
        if not self.is_NeuralSDE:
            return False
        return list(self.func.sigma.parameters())[-1].shape[0] == 1

    def __call__(self):

        if self.is_NeuralODE:
            if self._adjoint:
                self._integrator = torchdiffeq.odeint_adjoint
            else:
                self._integrator = torchdiffeq.odeint
            return "NeuralODE", self.mu_is_potential_func, self.sigma_is_potential_func

        if self.is_NeuralSDE:
            if self._adjoint:
                self._integrator = torchsde.sdeint_adjoint
            else:
                self._integrator = torchsde.sdeint
            return "NeuralSDE", self.mu_is_potential_func, self.sigma_is_potential_func

        if self.is_TorchNet:
            self._integrator = brownian_diffuser.nn_int
            return "TorchNet", self.mu_is_potential_func, self.sigma_is_potential_func

def function_credentials(func, adjoint=False):
    
    creds = Credentials(func, adjoint=adjoint)
    func_type, mu_is_potential, sigma_is_potential = creds()
    
    return {
        "func_type": func_type,
        "mu_is_potential": mu_is_potential,
        "sigma_is_potential": sigma_is_potential,
        "use_adjoint": adjoint,
    }



# -- import packages: ----------------------------------------------------------
from neural_diffeqs import NeuralSDE
import torch


# -- import local dependencies: ------------------------------------------------
from . import base, mix_ins


from typing import Union, List


# -- lightning model: ----------------------------------------------------------
class LightningSDE(
    mix_ins.BaseForwardMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        latent_dim,
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        sigma_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.1,
        sigma_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        train_lr=1e-4,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        *args,
        **kwargs,
    ):
        super().__init__()

        self.save_hyperparameters()
        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=NeuralSDE, kwargs=locals())
        self._configure_optimizers_schedulers()

    def __repr__(self):
        return "LightningSDE"




__module_name__ = "__init__.py"
__version__ = "0.0.45"
__doc__ = """plotting __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# import functions accessed as sdq.pl.<func>: --------------------------------------------




# -- import packages: --------------------------------------------------------------------
from abc import ABC
import inspect


# -- Main class: -------------------------------------------------------------------------
class AutoParseBase(ABC):

    def __init__(self):
        self._KWARGS = {}
    
    def __call__(self):
        pass
    
    def __init_kwargs__(self):
        self._KWARGS = {}
        
    def _inspect(self, func):
        return list(inspect.signature(func).parameters.keys())
        
    @property
    def _init_params(self):
        return self._inspect(func=self.__init__)
    
    @property
    def _call_params(self):
        return self._inspect(func=self.__call__)
    
    @property
    def _parse_params(self):
        return self._inspect(func=self.__parse__)

    @property
    def _collected_params(self):
        return self._init_params + self._call_params + self._parse_params        

    def _collect_literal_kwargs(self, kwargs_val):
        
        for key, val in kwargs_val.items():
            self.__collect__(key, val)
    
    def __hide__(self, key):
        return "_{}".format(key)

    def __collect__(self, key, val):
        if not hasattr(self, "_KWARGS"):
            self.__init_kwargs__()
            
        self._KWARGS[key] = val
        setattr(self, key, val)

    def __parse__(
        self,
        kwargs:  dict,
        ignore:  list = ["self"],
        private: list = ["ignore", "private", "public"],
        public:  list = [],
        kwargs_key: str = "kwargs",
    ):
        """
        Pass `locals()` or some other collection of kwargs to `kwargs` to
        save them as attributes of the subclass.
        
        Parameters:
        -----------
        kwargs
            typically `locals()`
            type: dict
        
        ignore
            type: list
            default: ["self"]
        
        private
            type: list
            default: ["ignore", "private", "public"]
        
        public
            type: list
            default: []
        
        kwargs_key
            type: str
            default: "kwargs"

        Notes:
        ------
        (1) assumes all are public unless denoted in private
        (2) If a public list is provided, all kwargs are shifted to private unless denoted in public.
        """

        if len(public) > 0:
            private = list(kwargs.keys())

        for key, val in kwargs.items():
            if not key in ignore: # (key in self._collected_params) and ():
                if key == kwargs_key:
                    self._collect_literal_kwargs(val)
                else:
                    if (key in private) and (not key in public):
                        key = self.__hide__(key)
                    self.__collect__(key, val)





# -- import packages: ----------------------------------------------------------
from tqdm.notebook import tqdm
import lightning
import anndata
import pandas as pd
import torch
import glob
import os


# -- import local dependencies: ------------------------------------------------
from . import configs, lightning_models, utils
from .. import tools


# -- type setting: -------------------------------------------------------------
from typing import Union, List
NoneType = type(None)


class scDiffEq(utils.ABCParse):
    def __init__(
        self,
        
        # -- data params: -------------------------------------------------------
        adata: anndata.AnnData,
        latent_dim: int = 20,
        model_name: str = "scDiffEq_model",
        use_key: str = "X_scaled",
        obs_keys: List[str] = ["W"],
        kNN_key: str = "X_pca_scDiffEq",
        
        # -- pretrain params: ---------------------------------------------------
        pretrain_epochs: int = 500,
        pretrain_lr: float = 1e-3,
        pretrain_optimizer=torch.optim.Adam,
        pretrain_step_size: int = 100,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        
        # -- train params: ------------------------------------------------------
        train_epochs: int =1500,
        train_lr: float = 1e-5,
        train_optimizer = torch.optim.RMSprop,
        train_scheduler = torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        train_val_split: List[float] = [0.9, 0.1],
        batch_size: int = 2000,
        
        train_key: str = "train",
        val_key: str = "val",
        test_key: str = "test",
        predict_key: str = "predict",
        
        # -- general params: ----------------------------------------------------
        seed: int = 617,
        num_workers: int = os.cpu_count(),
        silent: bool = True,
        scale_input_counts: bool = True,
        reduce_dimensions: bool = True,
        build_kNN: bool = True,
        fate_bias_csv_path: Union[str, NoneType] = None,
        fate_bias_multiplier: float = 1,
        
        # -- time params: -------------------------------------------------------
        time_key: Union[str, NoneType] = None,
        t0_idx: Union[pd.Index, NoneType] = None,
        t_min: float = 0,
        t_max: float = 1,
        dt: float = 0.1,
        time_cluster_key=None,
        t0_cluster=None,
        shuffle_time_labels = False,
        
        # -- DiffEq params: ----------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        
        sigma_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_dropout: Union[float, List[float]] = 0.1,
        sigma_bias: List[bool] = True,
        sigma_output_bias: bool = True,
        sigma_n_augment: int = 0,
        
        adjoint: bool = False,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        coef_prior_drift: float = 1.0,
        DiffEq_type: str = "SDE",
        potential_type: str = "prior",
        
        # -- Encoder params: ---------------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        
        # -- Decoder params: ---------------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        
        *args,
        **kwargs,
    ):
        self.__config__(locals())

    def _configure_obs_idx(self):
        if not self.adata.obs.index[0] == "1":
            self.adata = utils.idx_to_int_str(self.adata)
        
        
    def _configure_data(self, kwargs):

        """Configure data (including time time)"""
        
        self._configure_obs_idx()
        
        self.t, self.t_config = configs.configure_time(
            **utils.extract_func_kwargs(func = configs.configure_time, kwargs = kwargs),
        )
        kwargs['groupby'] = self.t_config.attributes['time_key']
        kwargs.update(self.t_config.attributes)
        self.LitDataModule = configs.LightningData(
            **utils.extract_func_kwargs(
                func=configs.LightningData,
                kwargs=kwargs,
            )
        )
        self._data_dim = self.LitDataModule.n_dim

    def _configure_dimension_reduction(self):
        self.reducer = tools.DimensionReduction(self.adata, save_path = self.DiffEqLogger.default_model_outdir)
        if self._scale_input_counts:
            self._INFO("Scaling input counts (for dimension reduction).")
            self.reducer.scale()
            self.adata.obsm["X_scaled_scDiffEq"] = self.reducer.X_scaled
        else:
            self.reducer.X_scaled_train = reducer.X_train
            self.reducer.X_scaled = reducer.X

        if self._reduce_dimensions:
            self.reducer.pca()
            self.adata.obsm["X_pca_scDiffEq"] = self.reducer.X_pca
            self.reducer.umap()
            self.adata.obsm["X_umap_scDiffEq"] = self.reducer.X_umap

    def _configure_logger(self):
        self.DiffEqLogger = utils.scDiffEqLogger(model_name=self._model_name)
        self.DiffEqLogger()

    def _configure_trainer_generator(self):
        
        self.TrainerGenerator = configs.LightningTrainerConfiguration(
            self.DiffEqLogger.versioned_model_outdir
        )
        self._PRETRAIN_CONFIG_COUNT = 0
        self._TRAIN_CONFIG_COUNT = 0

    def _configure_kNN_graph(self):
        train_adata = self.adata[self.adata.obs[self._train_key]].copy()
        
        train_adata.obs = train_adata.obs.reset_index(drop=True)
        train_adata.obs.index = train_adata.obs.index.astype(str)
        
        self._INFO(f"Bulding Annoy kNN Graph on adata.obsm['{self._kNN_key}']")
        self.kNN_Graph = utils.kNNGraphQuery(
            adata=train_adata, use_key=self._kNN_key,
        )

    def _configure_model(self, kwargs):

        self._LitModelConfig = configs.LightningModelConfiguration(
            data_dim=self._data_dim,
            latent_dim=self._latent_dim,
            DiffEq_type=self._DiffEq_type,
            potential_type=self._potential_type,
            fate_bias_csv_path = self._fate_bias_csv_path,
        )
        if hasattr(self, "reducer"):
            kwargs['PCA'] = self.reducer.PCA
        
        if hasattr(self, "kNN_Graph"):
            kwargs['kNN_Graph'] = self.kNN_Graph

        self.DiffEq = self._LitModelConfig(kwargs)
        self._INFO(f"Using the specified parameters, {self.DiffEq} has been called.")
    
    def __config__(self, kwargs):

        """
        Run on model.__init__()

        Step 0: Parse all kwargs
        Step 1: Set up info messaging [TODO: eventually replace this with more sophisticated logging]
        Step 2: Configure data
        Step 4: Configure dimension reduction models
        Step 5: Configure kNN Graph.
        Step 3: Configure lightning model
        Step 6: Configure logger
        Step 7: Configure TrainerGenerator
        """

        self.__parse__(kwargs, public=["adata"])
        self._INFO = utils.InfoMessage()
        self._configure_data(kwargs)
        self._configure_logger()
        if kwargs["reduce_dimensions"]:
            self._configure_dimension_reduction()
        if kwargs["build_kNN"]:
            self._configure_kNN_graph()
        self._configure_model(kwargs)
        self._configure_trainer_generator()

    def freeze(self):
        """Freeze lightning model"""
        self.DiffEq.freeze()

    def load(self, ckpt_path, freeze=True):
        self.ckpt_path = ckpt_path
        self.DiffEq.load_from_checkpoint(ckpt_path)
        if freeze:
            self.DiffEq.freeze()

    def _stage_log_path(self, stage):
        log_path = glob.glob(self.DiffEqLogger.versioned_model_outdir + f"/{stage}*")[0]
        self._INFO(f"Access logs at: {log_path}")
        
    @property
    def _VERSION(self):
        return int(os.path.basename(self.DiffEqLogger.versioned_model_outdir).split("_")[1])
    
    def _check_disable_validation(self, trainer_kwargs):
        if self._train_val_split[1] == 0:
            trainer_kwargs.update(
                {
                    'check_val_every_n_epoch': 0,
                    'limit_val_batches': 0.0,
                    'num_sanity_val_steps': 0.0,
                    'val_check_interval': 0.0,
                },

            )
            
        return trainer_kwargs
            
    def _configure_pretrain_step(self, epochs, callbacks=[]):
        
        STAGE = "pretrain"
        self._INFO(f"Configuring fit step: {STAGE}")
        
        self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)

#         if not isinstance(epochs, NoneType):
            
#             if self._PRETRAIN_CONFIG_COUNT > 0:
#                 epochs = epochs + self._pretrain_epochs
            
#             self._pretrain_epochs = epochs
#             self._PARAMS["pretrain_epochs"] = epochs
#             self.DiffEq.hparams['pretrain_epochs'] = epochs
#             self._INFO(f"Pretrain epochs scheduled: {epochs}")

        trainer_kwargs = utils.extract_func_kwargs(
            func=self.TrainerGenerator,
            kwargs=self._PARAMS,
        )
        trainer_kwargs.update(
            utils.extract_func_kwargs(
                func=lightning.Trainer,
                kwargs=self._PARAMS,
            )
        )
        trainer_kwargs.update(
            utils.extract_func_kwargs(
                func=lightning.Trainer,
                kwargs=locals(),
            )
        )
        trainer_kwargs = self._check_disable_validation(trainer_kwargs)
        
        self.pre_trainer = self.TrainerGenerator(
            max_epochs=self._pretrain_epochs,
            stage=STAGE,
            working_dir = self.DiffEqLogger.wd,
            version = self._VERSION,
            pretrain_version=self._PRETRAIN_CONFIG_COUNT,
            train_version=self._TRAIN_CONFIG_COUNT,
            **trainer_kwargs
        )
        self._stage_log_path(STAGE)

    def pretrain(
        self,
        epochs=None,
        callbacks = [],
    ):
        """If any of the keyword arguments are passed, they will replace the previously-stated arguments from __init__ and re-configure the DiffEq."""

        self._configure_pretrain_step(epochs, callbacks)
        self.pre_trainer.fit(self.DiffEq, self.LitDataModule)

    def _configure_train_step(self, epochs, kwargs):
                    
        STAGE = "train"

        self._INFO(f"Configuring fit step: {STAGE}")
        
        self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)

#         if not isinstance(epochs, NoneType):
#             if self._TRAIN_CONFIG_COUNT > 0:
#                 epochs = epochs + self._train_epochs
                
#             self._train_epochs = epochs
#             self._PARAMS["train_epochs"] = epochs
#             self.DiffEq.hparams['train_epochs'] = epochs
#             self._INFO(f"Train epochs scheduled: {epochs}")

        trainer_kwargs = utils.extract_func_kwargs(
            func=self.TrainerGenerator,
            kwargs=self._PARAMS,
        )
        trainer_kwargs.update(
            utils.extract_func_kwargs(
                func=lightning.Trainer,
                kwargs=self._PARAMS,
            )
        )
        trainer_kwargs.update(
            utils.extract_func_kwargs(
                func=lightning.Trainer,
                kwargs=kwargs,
            )
        )
        
        trainer_kwargs = self._check_disable_validation(trainer_kwargs)
        
        self.trainer = self.TrainerGenerator(
            max_epochs=self._train_epochs,
            stage=STAGE,
            working_dir = self.DiffEqLogger.wd,
            version = self._VERSION,
            pretrain_version=self._PRETRAIN_CONFIG_COUNT,
            train_version=self._TRAIN_CONFIG_COUNT,
            **trainer_kwargs,
        )

        self._stage_log_path(STAGE)

    def train(
        self,
        epochs=500,
        callbacks=[],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor=None,
        accelerator=None,
        log_every_n_steps=1,
        reload_dataloaders_every_n_epochs=1,
        devices=None,
        deterministic=False,
        **kwargs,
    ):
        
        self.DiffEq._update_lit_diffeq_hparams(self._PARAMS)
        self._configure_train_step(epochs, locals())
        self.trainer.fit(self.DiffEq, self.LitDataModule)

    def fit(
        self,
        train_epochs=200,
        pretrain_epochs=500,
        train_lr = None,
        callbacks: List = [],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor=None,
        accelerator=None,
        log_every_n_steps=1,
        reload_dataloaders_every_n_epochs=1,
        devices=None,
        deterministic=False,
        **kwargs,
    ):
                
        if pretrain_epochs > 0 and (not self._LitModelConfig.use_vae):
            pretrain_epochs = 0
            
            
        self.__update__(locals())

        if pretrain_epochs > 0:
            self.pretrain(
                epochs=pretrain_epochs, **utils.extract_func_kwargs(self.pretrain, locals()))
        if train_epochs > 0:
            self.train(
                epochs=train_epochs, **utils.extract_func_kwargs(self.train, locals())
            )
            
        # TO-DO: eventually replace how this works...
        self._PRETRAIN_CONFIG_COUNT += 1
        self._TRAIN_CONFIG_COUNT += 1
        
        
    @property
    def loss(self):
        # scDiffEq_fit_loss_tracking.png
        utils.display_tracked_loss(self.DiffEqLogger)
        



# -- import packages: --------------------------------------------------------------------
import anndata
import pandas as pd
import numpy as np
import torch


# -- import local dependencies: ----------------------------------------------------------
from .. import utils
from ... import tools


# -- define types: -----------------------------------------------------------------------
NoneType = type(None)


class TimeConfiguration(utils.ABCParse):
    def __init__(
        self,
        adata: anndata.AnnData,
        time_key: str = None,
        t_min=0,
        t_max=1,
        dt=0.1,
        t0_idx=None,
        t0_cluster=None,
        time_cluster_key=None,
    ):
        super().__init__()
        self.__parse__(kwargs=locals(), public=["adata"])
        
## WE DONT NEED THIS B/C WE CAN GENERATE T0_IDX FROM OTHER KEY:VAL PAIRS
#         self._check_passed_time_args(self._t0_idx, self._time_key)
        
#     def _check_passed_time_args(self, t0_idx, time_key):
#         """If time_key is passed"""
        
#         if sum([utils.not_none(time_key), utils.not_none(t0_idx)]) < 1:
#             raise ValueError("Must provide `t0_idx` and/or `time_key`.")

    @property
    def t0_idx(self):

        """
        If t0_idx is provided, returns that.
        If t0_idx is not provided, but a cluster key and t0 cluster
        value-pair is provided, this can isolate the t0_idx.
        """

        if self.has_time_key:
            
            return self.adata.obs[
                self.adata.obs[self._time_key] == self.adata.obs[self._time_key].min()
            ].index

        if not isinstance(self._t0_idx, NoneType):
            return self._t0_idx

        if (not isinstance(self._time_cluster_key, NoneType)) and (
            not isinstance(self._t0_cluster, NoneType)
        ):
            return self.adata.obs[
                self.adata.obs[self._time_cluster_key] == self._t0_cluster
            ].index

    def time_sampling(self):
        if not self.has_time_key:
            self.adata.obs["t0"] = self.adata.obs.index.isin(self.t0_idx)
#         else:
        tools.time_free_sampling(
            adata=self.adata,
            t0_idx=self.t0_idx,
            n_steps=self.n_steps,
            t0_key="t0",
            t_key="t",
        )

    @property
    def has_time_key(self):
        return (not isinstance(self._time_key, NoneType)) and (
            self._time_key in self.adata.obs_keys()
        )

    @property
    def time_key(self):
        if self.has_time_key:
            return self._time_key
        return "t"

    @property
    def t_min(self):
        if self.has_time_key:
            self._t_min = self.adata.obs[self._time_key].min()
        return self._t_min

    @property
    def t_max(self):
        if self.has_time_key:
            self._t_max = self.adata.obs[self._time_key].max()
        return self._t_max

    @property
    def t_diff(self):
        return self._t_max - self._t_min

    def _IS_CONTIGUOUS(self, n_steps):
        return n_steps % 1 == 0

    @property
    def n_steps(self):
        _n_steps = self.t_diff / self._dt + 1
        if self._IS_CONTIGUOUS(_n_steps):
            return int(_n_steps)
        raise ValueError("Time is non-contiguous")

    @property
    def dt(self):
        return self._dt

    def _t_from_time_key(self):
        return torch.Tensor(sorted(self.adata.obs[self._time_key].unique()))

    def _t_from_bounds(self):
        return torch.linspace(self._t_min, self._t_max, self.n_steps)

    def __call__(self):

        if self.has_time_key:
            return self._t_from_time_key()
        self.time_sampling()
        return self._t_from_bounds()

    @property
    def attributes(self):
        self._ATTR_KEYS = sorted(
            ["time_key", "t0_idx", "t_min", "t_max", "t_diff", "dt", "n_steps"]
        )
        return {attr: getattr(self, attr) for attr in self._ATTR_KEYS}

    def __repr__(self):

        header = "Time Attributes:\n\n  "

        _attrs = [
            "{:<9} :  {}".format(attr, val) for attr, val in self.attributes.items()
        ]
        return header + "\n  ".join(_attrs)


def configure_time(
    adata: anndata.AnnData,
    time_key: str = None,
    t_min: float = 0,
    t_max: float = 1,
    t0_idx: pd.Index = None,
    dt: float = 0.1,
    t0_cluster=None,
    time_cluster_key=None,
) -> dict:

    """Updates time for AnnData. Returns time_attributes dictionary."""

    time_config = TimeConfiguration(
        adata=adata,
        time_key=time_key,
        t_min=t_min,
        t_max=t_max,
        t0_idx=t0_idx,
        dt=dt,
        t0_cluster=t0_cluster,
        time_cluster_key=time_cluster_key,
    )
    
    t = time_config()
    
    return t, time_config

# # -- main class: -------------------------------------------------------------------------
# class TimeConfiguration(utils.AutoParseBase):
#     def __init__(
#         self,
#         adata: AnnData,
#         time_key: str = None,
#         t0_idx: pd.Index = None,
#         t: torch.Tensor = None,
#         dt: float = None,
#         n_steps: int = 40,
#         t_min: float = 0,
#         t_max: float = 1,
#     ):
#         """
#         Must provide t0_idx or time_key. if both are provided,
#         time_key overrules t0_idx.
#         """
#         self._check_passed_time_args(t0_idx, time_key)
#         self.__parse__(locals(), public=[None])

#     # -- utility funcs: ------------------------------------------------------------------
#     def _check_passed_time_args(self, t0_idx, time_key):
#         """If time_key is passed"""
#         if utils.not_none(t0_idx):
#             time_key = None
#         elif sum([utils.not_none(time_key), utils.not_none(t0_idx)]) < 1:
#             "Must provide t0_idx and/or time_key. If both are provided, t0_idx overrules time_key"
            
    
# #     def _value_pass_check(self, time_key, t0_idx):
# #         assert any(
# #             [not isinstance(time_key, NoneType), not isinstance(t0_idx, NoneType)]
# #         ), "Must provide t0_idx or time_key. If both are provided, time_key overrules t0_idx"

#     def _time_from_adata(self):
#         return torch.Tensor(sorted(self._adata.obs[self.time_key].unique()))

#     # -- time_key: -----------------------------------------------------------------------
#     def _configure_time_key(self):
#         if isinstance(self._time_key, NoneType):
#             try:

#                 time_free_sampling(
#                     self._adata,
#                     self.t0_idx,
#                     n_steps=self.n_steps,
#                     t0_key="t0",
#                     t_key="t",
#                 )
#                 utils.normalize_time(
#                     self._adata,
#                     time_key="t",
#                     t_min=self._t_min,
#                     t_max=self._t_max,
#                 )
#                 self._time_key = "t"

#             except:
#                 raise ValueError("Pass `t0_idx` or `time_key`")

#     @property
#     def time_key(self):
#         self._configure_time_key()
#         return self._time_key

#     # -- t: ------------------------------------------------------------------------------
#     def _configure_t(self):
#         """
#         we are here because we don't have t
#         we can use 0-1 in place of a real time and throw in n_steps
#         """

#         if isinstance(self._t, NoneType):
#             self._t = self._time_from_adata()

#     @property
#     def t(self):
#         self._configure_t()
#         return self._t

#     # -- t0_idx: -------------------------------------------------------------------------
#     def _configure_t0_idx(self):
#         if isinstance(self._t0_idx, NoneType):
#             raise ValueError("Must provide t0_idx")

#     @property
#     def t0_idx(self):
#         self._configure_t0_idx()
#         return self._t0_idx

#     # -- t_min: --------------------------------------------------------------------------
#     def _configure_t_min(self):
#         if not isinstance(self.t, NoneType):
#             self._t_min = min(self.t).item()

#     @property
#     def t_min(self) -> float:
#         self._configure_t_min()
#         return self._t_min

#     # -- t_max: --------------------------------------------------------------------------
#     def _configure_t_max(self):
#         if not isinstance(self.t, NoneType):
#             self._t_max = max(self.t).item()

#     @property
#     def t_max(self) -> float:
#         self._configure_t_max()
#         return self._t_max

#     # -- t_span: -------------------------------------------------------------------------
#     def _configure_t_span(self):
#         if not hasattr(self, "_t_span"):
#             self._t_span = abs(self.t_max - self.t_min)

#     @property
#     def t_span(self):
#         self._configure_t_span()
#         return self._t_span

#     # -- dt: -----------------------------------------------------------------------------
#     def _configure_dt(self):
#         """
#         we are here because we don't have dt.
#         to get dt, de novo, we need t and n_steps.
#         """
#         if isinstance(self._dt, NoneType):
#             self._dt = self.t_span / self._n_steps

#     @property
#     def dt(self):
#         self._configure_dt()
#         return self._dt

#     # -- n_steps: ------------------------------------------------------------------------
#     @property
#     def n_steps(self):
#         return int(self._n_steps + 1)

#     # -- collect attributes: -------------------------------------------------------------

#     def __attributes__(self):

#         self._attrs = {}

#         for attr in self.__dir__():
#             if (not attr == "attributes") and (not attr.startswith("_")):
#                 if attr == "t0_idx":
#                     if self._t0_idx is None:
#                         continue
#                 self._attrs[attr] = getattr(self, attr)

#         return self._attrs

#     @property
#     def attributes(self):
#         return self.__attributes__()
    
# def configure_time(
#     adata: AnnData,
#     time_key: str = None,
#     t0_idx: pd.Index = None,
#     t: torch.Tensor = None,
#     dt: float = None,
#     n_steps: int = 40,
#     t_min: float = 0,
#     t_max: float = 1,
# ) -> dict:

#     """Updates time for AnnData. Returns time_attributes dictionary."""

#     time_config = TimeConfiguration(
#         adata=adata,
#         time_key=time_key,
#         t0_idx=t0_idx,
#         t=t,
#         dt=dt,
#         n_steps=n_steps,
#         t_min=t_min,
#         t_max=t_max,
#     )
#     return time_config.attributes




__module_name__ = "__init__.py"
__version__ = "0.0.46rc2"
__doc__ = """Top-level __init__ for accessing the scDiffEq model."""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# -- import models accessed as sdq.models.<MODEL>: ---------------------------------------
from ._scdiffeq import scDiffEq

from . import lightning_models
from . import utils
from . import configs
from . import callbacks



from abc import ABC
from typing import Dict, List, Any
NoneType = type(None)

class ABCParse(ABC):
    _BUILT = False

    def __init__(self, *args, **kwargs):
        """
        we avoid defining things in __init__ because this subsequently
        mandates the use of `super().__init__()`
        """
        pass

    def __build__(self) -> None:

        self._PARAMS = {}
        self._IGNORE = ["self", "__class__"]
        self._stored_private = []
        self._stored_public = []

        self._BUILT = True

    def __set__(
        self, key: str, val: Any, public: List = [], private: List = []
    ) -> None:

        self._PARAMS[key] = val
        if (key in private) and (not key in public):
            self._stored_private.append(key)
            key = f"_{key}"
        else:
            self._stored_public.append(key)
        setattr(self, key, val)

    def __set_existing__(self, key: str, val: Any) -> None:

        self._PARAMS[key] = val

        if key in self._stored_private:
            key = f"_{key}"
        setattr(self, key, val)

    @property
    def _STORED(self) -> List:
        return self._stored_private + self._stored_public

    def __parse__(
        self, kwargs: Dict, public: List = [], private: List = [], ignore: List = []
    ):

        if not self._BUILT:
            self.__build__()

        self._IGNORE += ignore

        if len(public) > 0:
            private = list(kwargs.keys())

        for key, val in kwargs.items():
            if not key in self._IGNORE:
                self.__set__(key, val, public, private)

    def __update__(self, kwargs: dict, public: List = [], private: List = []) -> None:

        if not self._BUILT:
            self.__build__()

        for key, val in kwargs.items():
            if not isinstance(val, NoneType) and key in self._STORED:
                self.__set_existing__(key, val)
            elif not isinstance(val, NoneType) and not key in self._IGNORE:
                self.__set__(key, val, public, private)




__module_name__ = "__init__.py"
__doc__ = """Main __init__ module - most user-visible API."""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(["mvinyard@broadinstitute.org", "arasmuss@broadinstitute.org", "ruitong@broadinstitute.org"])


# specify version: -----------------------------------------------------------------------
__version__ = "0.0.44"


# import packages: -----------------------------------------------------------------------
import anndata


# API-facing function: -------------------------------------------------------------------
def read_h5ad(h5ad_path: str, silent: bool=False)->anndata.AnnData:

    """
    Read an AnnData object from a path to a .h5ad file.

    Parameters:
    -----------
    h5ad_path [ required ]
        path to .h5ad file.
        type: str


    silent [ optional ]
        If True, the anndata object is not printed.
        default: False
        type: bool

    Returns:
    --------
    adata
        type: anndata._core.anndata.AnnData

    Notes:
    ------
    (1) Documentation for AnnData: https://anndata.readthedocs.io/en/stable/

    """

    adata = anndata.read_h5ad(h5ad_path)

    if not silent:
        print(adata)

    return adata


from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from umap import UMAP
import ABCParse
import pickle
import os

from ..core import utils



class DimensionReduction(ABCParse.ABCParse):
    def __init__(
        self,
        adata,
        n_pcs=50,
        n_umap_components=2,
        train_key="train",
        test_key="test",
        write=True,
        save_path="./",
        pkl_protocol=4,
        use_saved=True,
    ):

        self.__parse__(locals(), private=[None])

        self.df = self.adata.obs.copy()
        self._INFO = utils.InfoMessage()

        self.train_idx = self.adata[self.df[self.train_key]].obs.index
        self.test_idx = self.adata[self.df[self.test_key]].obs.index
        self.unique_test_idx = self.test_idx[
            [idx not in self.train_idx for idx in self.test_idx]
        ]

        self.X_train = self.adata[self.train_idx].X.toarray()
        self.X = self.adata.X
        
        self.n_cells_train = self.X_train.shape[0]
        self.n_cells  = self.X.shape[0]

        self.SCALER_KWARGS = utils.function_kwargs(
            func=StandardScaler, kwargs=self._PARAMS
        )
        self.PCA_KWARGS = utils.function_kwargs(func=PCA, kwargs=self._PARAMS)
        self.UMAP_KWARGS = utils.function_kwargs(func=UMAP, kwargs=self._PARAMS)

    @property
    def scaler(self):
        if not hasattr(self, "_scaler_model"):
            if os.path.exists(self.path_scaler) and (not self.use_saved):
                self._scaler_model = pickle.load(open(self.path_scaler, "rb"))
            else:
                self._scaler_model = StandardScaler(**self.SCALER_KWARGS)
        return self._scaler_model

    @property
    def PCA(self):
        if not hasattr(self, "_pca_model"):
            if os.path.exists(self.path_pca) and self.use_saved:
                self._pca_model = pickle.load(open(self.path_pca, "rb"))
            else:
                self._INFO(f"Fitting PCA model on {self.n_cells_train} training cells")
                self._pca_model = PCA(n_components=self.n_pcs, **self.PCA_KWARGS)
        return self._pca_model

    @property
    def UMAP(self):
        if not hasattr(self, "_UMAP_model"):
            if os.path.exists(self.path_umap) and self.use_saved:
                self._INFO("Loading UMAP model from file")
                self._UMAP_model = pickle.load(open(self.path_umap, "rb"))
            else:
                self._UMAP_model = UMAP(
                    n_components=self.n_umap_components, **self.UMAP_KWARGS
                )
        return self._UMAP_model

    @property
    def path_scaler(self):
        return os.path.join(self.save_path, "scaler_model.pkl")

    @property
    def path_pca(self):
        return os.path.join(self.save_path, "pca_model.pkl")

    @property
    def path_umap(self):
        return os.path.join(self.save_path, "umap_model.pkl")

    def scale(self):
        self.X_scaled_train = self.scaler.fit_transform(self.X_train)
        self.X_scaled = self.scaler.transform(self.X)
        if self.write:
            pickle.dump(
                obj=self.scaler,
                file=open(self.path_scaler, "wb"),
                protocol=self.pkl_protocol,
            )

    def pca(self):
        self.X_pca_train = self.PCA.fit_transform(self.X_scaled_train)
        self._INFO(f"Transforming all cells ({self.n_cells}) using trained PCA model")
        self.X_pca = self.PCA.transform(self.X_scaled)
        if self.write:
            self._INFO(f"Saving PCA model to: {self.path_pca}")
            pickle.dump(
                obj=self.PCA,
                file=open(self.path_pca, "wb"),
                protocol=self.pkl_protocol,
            )

    def umap(self):
        if os.path.exists(self.path_umap) and self.use_saved:
            self.X_umap_train = self.UMAP.embedding_
            self._INFO(f"Transforming all cells ({self.n_cells}) using trained UMAP model")
            self.X_umap = self.UMAP.transform(self.X_pca)
        else:
            self._INFO(f"Fitting UMAP model on {self.n_cells_train} training cells")
            self.X_umap_train = self.UMAP.fit_transform(self.X_pca_train)
            self._INFO(f"Transforming all cells ({self.n_cells}) using trained UMAP model")
            self.X_umap = self.UMAP.transform(self.X_pca)
            if self.write:
                self._INFO(f"Saving UMAP model to: {self.path_umap}")
                pickle.dump(
                    obj=self.UMAP,
                    file=open(self.path_umap, "wb"),
                    protocol=self.pkl_protocol,
                )



__module_name__ = "__init__.py"
__version__ = "0.0.44"
__doc__ = """tools __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# import functions accessed as sdq.tl.<func>: --------------------------------------------
from ._annotate_cells import annotate_cells
from ._time_free_sampling import time_free_sampling
from ._hyperparams import HyperParams
from ._reconstruct_function import reconstruct_function
from ._versions import Versions, configure_version
from ._func_from_version import func_from_version
from ._umap import UMAP
from ._fetch import fetch
from ._drift_diffusion_state_characterization import drift, diffusion

from ._dimension_reduction import DimensionReduction



from ... import tools

from autodevice import AutoDevice
NoneType = type(None)

def fetch_format(adata, use_key: str, idx=None, N=1, device=AutoDevice()):

    if isinstance(idx, NoneType):
        idx = range(len(adata))

    return tools.fetch(adata[idx], use_key=use_key, device=device)[:, None].expand(
        -1, N, -1
    )




import licorice_font

class InfoMessage:
    def __init__(self, INFO="INFO", color="BLUE"):
        self.INFO = licorice_font.font_format(INFO, [color])

    def _format_msg_title(self, INFO):
        return f" - [{INFO}] | "

    def __call__(self, msg):

        self.BASE = self._format_msg_title(self.INFO)
        print(self.BASE + msg)



# -- import packages: ------
import brownian_diffuser
import neural_diffeqs
import torchdiffeq
import torchsde
import torch


# -- import local dependencies: ------
from .. import lightning_models, utils


# -- supporting functions: ------------------------------------
def fetch_optimizer(func):
    """
    Examples:
    ---------
    fetch_optimizer("RMSprop")
    >>> torch.optim.rmsprop.RMSprop

    fetch_optimizer(torch.optim.RMSprop)
    >>> torch.optim.rmsprop.RMSprop
    """
    fetch = utils.FunctionFetch(module=torch.optim, parent=torch.optim.Optimizer)
    return fetch(func)


def fetch_lr_scheduler(func):
    """
    Examples:
    ---------
    fetch_lr_scheduler("StepLR")
    >>> torch.optim.lr_scheduler.StepLR

    fetch_lr_scheduler(torch.optim.lr_scheduler.StepLR)
    >>> torch.optim.lr_scheduler.StepLR
    """
    fetch = utils.FunctionFetch(
        module=torch.optim.lr_scheduler, parent=torch.optim.lr_scheduler._LRScheduler
    )
    return fetch(func)


# class ConfigureLightningModel:
class ChooseLitModel:
    def __init__(self, data_dim, latent_dim, de_type="SDE", potential_type=None):
        """
        data_dim
            [ required ]

        latent_dim
            [ required ]

        de_type
            options: ["SDE", "ODE"]
            default: "ODE"

        potential_type
            options: [None, "fixed", "prior"]
            default: None
        """

        self.data_dim = data_dim
        self.latent_dim = latent_dim
        self._de_type = de_type
        self._potential_type = potential_type

    @property
    def de_type(self):
        return "".join(["Lightning", self._de_type])

    @property
    def potential_type(self):
        potential_types = {
            "fixed": "FixedPotential",
            "prior": "PriorPotential",
        }
        if self._potential_type:
            return potential_types[self._potential_type]

    @property
    def use_vae(self):
        return self.data_dim > self.latent_dim

    @property
    def model(self):

        _model = [self.de_type]

        if self.use_vae:
            _model.append("VAE")

        if self.potential_type:
            _model.append(self.potential_type)

        return "_".join(_model)


# -- main operator class: ------------------------------------------------------------
class LightningModelConfiguration:
    def __init__(
        self,
        data_dim,
        latent_dim,
        de_type="SDE",
        potential_type=None,
#         func,
        optimizer=torch.optim.RMSprop,
        lr_scheduler=torch.optim.lr_scheduler.StepLR,
        adjoint=False,
    ):
        
        self._lit_model_choice = ChooseLitModel(
            data_dim=data_dim,
            latent_dim=latent_dim,
            de_type=de_type,
            potential_type=potential_type,
        )
#         self.func = func
        self._adjoint = adjoint
        self._optimizer = optimizer
        self._lr_scheduler = lr_scheduler

#         self.func_type, self.mu_is_potential, self.sigma_is_potential = self.function_credentials

#     @property
#     def function_credentials(self):
#         if isinstance(self.func, neural_diffeqs.NeuralODE):
#             if self._adjoint:
#                 integrator = torchdiffeq.odeint_adjoint
#             else:
#                 integrator = torchdiffeq.odeint
#             return "NeuralODE", self._is_mu_potential(), self._is_sigma_potential()

#         if isinstance(self.func, neural_diffeqs.NeuralSDE):
#             if self._adjoint:
#                 integrator = torchsde.sdeint_adjoint
#             else:
#                 integrator = torchsde.sdeint
#             return "NeuralSDE", self._is_mu_potential(), self._is_sigma_potential()

#         integrator = brownian_diffuser.nn_int
#         self._integrator = brownian_diffuser.nn_int
#         return "DriftNet", self._is_mu_potential(), self._is_sigma_potential()
    
#     def _is_mu_potential(self):
#         """Assumes potential is 1-D"""
# #         if not isinstance(self.func, neural_diffeqs.NeuralODE):
# #             return False
#         return list(self.func.mu.parameters())[-1].shape[0] == 1

#     def _is_sigma_potential(self):
#         """Assumes potential is 1-D"""
#         if not isinstance(self.func, neural_diffeqs.NeuralSDE):
#             return False
#         return list(self.func.sigma.parameters())[-1].shape[0] == 1

    def _non_import(self, attr):
        return (
            attr.startswith("_")
            or attr.startswith("Base")
            or (attr in ["SinkhornDivergence"])
        )

    @property
    def available_models(self):
        """Available, implemented models"""
        return {
            attr: getattr(lightning_models, attr)
            for attr in lightning_models.__dir__()
            if not self._non_import(attr)
        }

    @property
    def FunctionModuleName(self):
        func_type = self.func_type.strip("Neural")
        if self.mu_is_potential:
            return "".join(["LightningPotential"] + [func_type])
        return "".join(["Lightning"] + [func_type])
    
    @property
    def optimizer(self):
        return fetch_optimizer(self._optimizer)
    
    @property
    def lr_scheduler(self):
        return fetch_lr_scheduler(self._lr_scheduler)


    def __call__(self, kwargs):
        """Return the LightningModel"""
        model = self.available_models[self.FunctionModuleName]
        MODEL_KWARGS = utils.extract_func_kwargs(func = model, kwargs = kwargs, ignore=['optimizer', 'lr_scheduler'])
        return model(
            func=self.func,
            optimizer = self.optimizer,
            lr_scheduler = self.lr_scheduler,
            **MODEL_KWARGS,
        )





from ._lightning_callbacks_configuration import LightningCallbacksConfiguration
from ._lightning_trainer_configuration import LightningTrainerConfiguration
from ._lightning_data_configuration import LightningData
from ._function_credentialling import function_credentials
from ._time_configuration import TimeConfiguration, configure_time

from ._lightning_model_configuration import LightningModelConfiguration
from ._lightning_model_configuration_v2 import LightningModelConfiguration



# -- import packages: -----------------------------------
from torch_adata import LightningAnnDataModule
import os


# -- import local dependencies: -------------------------
from .. import utils


class LightningData(LightningAnnDataModule, utils.AutoParseBase):
    def __init__(
        self,
        adata=None,
        h5ad_path=None,
        batch_size=2000,
        num_workers=os.cpu_count(),
        train_val_split=[0.8, 0.2],
        use_key="X_pca",
        obs_keys=[],
        weight_key='W',
        groupby="Time point",  # TODO: make optional
        train_key="train",
        val_key="val",
        test_key="test",
        predict_key="predict",
        silent=True,
        shuffle_time_labels = False,
        **kwargs,
    ):
        super(LightningData, self).__init__()
        
                    
        self.__parse__(locals(), public=[None])
        self._format_sinkhorn_weight_key()
        self._format_train_test_exposed_data()
        self.configure_train_val_split()
        

    @property
    def n_dim(self):
        if not hasattr(self, "_n_dim"):
            self._n_dim = self.train_dataset.X.shape[-1]
        return self._n_dim
    
#     def shuffle_t(self):
        
#         df = self._adata.obs.copy()
#         non_t0 = df.loc[df['t'] != 0]['t']
        
#         shuffled_t = np.zeros(len(df))
#         shuffled_t[non_t0.index.astype(int)] = np.random.choice(non_t0.values, len(non_t0))
#         self._adata.obs["t"] = shuffled_t
    
    def _format_sinkhorn_weight_key(self):
        if not self._weight_key in self._adata.obs.columns:
            self._adata.obs[self._weight_key] = 1
        self._obs_keys.append(self._weight_key)
    
    def _format_train_test_exposed_data(self):
        
        if not self._train_key in self._adata.obs.columns:
            self._adata.obs[self._train_key] = True
        if not self._test_key in self._adata.obs.columns:
            self._adata.obs[self._test_key] = False
            
    def prepare_data(self):
        ...

    def setup(self, stage=None):
        ...




# -- import packages: ----------------------------------------------------------
from geomloss import SamplesLoss
from autodevice import AutoDevice


# -- import local dependencies: ------------------------------------------------
from ...utils import AutoParseBase


# -- API-facing class: ---------------------------------------------------------
class SinkhornDivergence(SamplesLoss, AutoParseBase):
    def __init__(
        self,
        device=AutoDevice(),
        loss="sinkhorn",
        backend="online", # "auto",
#         diameter=,
        p=2,
        blur=0.1,
        scaling=0.7,
        debias=True,
        sample_axis=1,
        **kwargs
    ):
        super(SinkhornDivergence, self).__init__()

        self.__parse__(locals())




# -- import packages: -------------
from autodevice import AutoDevice
import torch_adata

def fetch(adata, use_key: str, device=AutoDevice()):
    return torch_adata.tl.fetch(adata, use_key).to(device)



class BaseForwardMixIn(object):
    def __init__(self):
        super().__init__()


    def forward(self, Z0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        return self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_sinkhorn_divergence(sinkhorn_loss, t=batch.t, stage=stage)




__module_name__ = "__init__.py"
__version__ = "0.0.45"
__doc__ = """TODO"""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# -- import: -----------------------------------------------------------------------------
from ._loss_accounting import LossAccounting
from ._intermittent_saves import IntermittentSaves
from ._gradient_potential_callback import GradientPotentialTest
from ._testing import Testing

from ._visualize_tracked_loss import VisualizeTrackedLoss
from ._visualize_predictions import VisualizePredictions



# -- import packages: ----------------------------------------------------------
import torch_nets
import lightning
import torch


from ... import utils

# -- set typing: ---------------------------------------------------------------
from typing import Union, List
NoneType = type(None)


# -- MixIn class: --------------------------------------------------------------        
#     def _configure_pretrain_optimizers(
#         self,
#         pretrain_lr=1e-3,
#         pretrain_step_size=10,
#         optimizer=torch.optim.Adam,
#         scheduler=torch.optim.lr_scheduler.StepLR,
#     ):

#         pretrain_optimizer = optimizer(
#             list(self.Encoder.parameters()) + list(self.Decoder.parameters()),
#             lr=pretrain_lr,
#         )

#         pretrain_scheduler = torch.optim.lr_scheduler.StepLR(
#             pretrain_optimizer, step_size=pretrain_step_size
#         )
#     def _configure_pretrain(
#         self,
#         pretrain_epochs=0,
#         pretrain_lr=1e-3,
#         pretrain_step_size=10,
#         optimizer=torch.optim.Adam,
#         scheduler=torch.optim.lr_scheduler.StepLR,
#     ):
#         self.pretrain_epochs = pretrain_epochs
        
#         # -- Reconstruction Loss Function: -------------------------------------
#         self.MSE_RL = torch.nn.MSELoss(reduction="sum")

        
#         self._configure_pretrain_optimizers(
#             pretrain_lr=pretrain_lr,
#             pretrain_step_size=pretrain_step_size,
#             optimizer=optimizer,
#             scheduler=scheduler,
#         )
        
#         # -- other necessities: ------------------------------------------------
#         self.automatic_optimization = False
        
#     def configure(
#         self,
#         data_dim: int,
#         latent_dim: int,
#         pretrain_epochs=0,
#         pretrain_lr=1e-3,
#         pretrain_step_size=10,
#         optimizer=torch.optim.Adam,
#         scheduler=torch.optim.lr_scheduler.StepLR,
#         encoder_n_hidden: int = 3,
#         encoder_power: float = 2,
#         encoder_activation: Union[str, List[str]] = "LeakyReLU",
#         encoder_dropout: Union[float, List[float]] = 0.2,
#         encoder_bias: bool = True,
#         encoder_output_bias: bool = True,
#         encoder: Union[NoneType, torch.nn.Sequential] = None,
#         decoder_n_hidden: int = 3,
#         decoder_power: float = 2,
#         decoder_activation: Union[str, List[str]] = "LeakyReLU",
#         decoder_dropout: Union[float, List[float]] = 0.2,
#         decoder_bias: bool = True,
#         decoder_output_bias: bool = True,
#         decoder: Union[NoneType, torch.nn.Sequential] = None,
#         mu_hidden: Union[List[int], int] = [2000, 2000],
#         sigma_hidden: Union[List[int], int] = [400, 400],
#         mu_activation: Union[str, List[str]] = 'LeakyReLU',
#         sigma_activation: Union[str, List[str]] = 'LeakyReLU',
#         mu_dropout: Union[float, List[float]] = 0.2,
#         sigma_dropout: Union[float, List[float]] = 0.2,
#         mu_bias: bool = True,
#         sigma_bias: List[bool] = True,
#         mu_output_bias: bool = True,
#         sigma_output_bias: bool = True,
#         mu_n_augment: int = 0,
#         sigma_n_augment: int = 0,
#         sde_type='ito',
#         noise_type='general',
#         brownian_dim=1,
#         coef_drift: float = 1.0,
#         coef_diffusion: float = 1.0,
#         coef_prior_drift: float = 1.0,
#     ):
        
#         KWARGS = {}
        
#         KWARGS["Encoder"] = utils.function_kwargs(func=torch_nets.Encoder, kwargs=locals())
#         KWARGS["SDE"] = utils.function_kwargs(func=self._configure_sde, kwargs=locals())
#         KWARGS["Decoder"] = utils.function_kwargs(func=torch_nets.Decoder, kwargs=locals())
        
#         if isinstance(encoder, NoneType):
#             self.encoder = torch_nets.Encoder(**KWARGS['encoder'])
#         else:
#             self.encoder = encoder
            
#         self._configure_sde()
        
#         if isinstance(decoder, NoneType):
#             decoder = torch_nets.Decoder(**KWARGS['decoder'])
#         else:
#             self.decoder = decoder
        
#         self._configure_SDE(
#             state_size=self.hparams["latent_dim"],
#             **utils.function_kwargs(
#                 func=self._configure_SDE,
#                 kwargs=locals(),
#                 ignore=['state_size'],
#         )

#         self._configure_pretrain(
#             **utils.function_kwargs(
#                 func=self._configure_pretrain,
#                 kwargs=locals(),
#             )

#         self._configure_integrator()


class VAEMixIn(object):
    def __init__(self, *args, **kwargs):
        super().__init__()

        # -- loss functions: ---------------------------------------------------
        self.reconstruction_loss = torch.nn.MSELoss(reduction="sum", **kwargs)
        
    def _configure_encoder(self):
        self.Encoder = torch_nets.Encoder(
            data_dim = self.hparams['data_dim'],
            latent_dim = self.hparams['latent_dim'],
            n_hidden=self.hparams["encoder_n_hidden"],
            power=self.hparams["encoder_power"],
            activation=self.hparams["encoder_activation"],
            dropout=self.hparams["encoder_dropout"],
            bias=self.hparams["encoder_bias"],
            output_bias=self.hparams["encoder_output_bias"],
        )

    def _configure_decoder(self):
        self.Decoder = torch_nets.Decoder(
            data_dim = self.hparams['data_dim'],
            latent_dim = self.hparams['latent_dim'],
            n_hidden=self.hparams["decoder_n_hidden"],
            power=self.hparams["decoder_power"],
            activation=self.hparams["decoder_activation"],
            dropout=self.hparams["decoder_dropout"],
            bias=self.hparams["decoder_bias"],
            output_bias=self.hparams["decoder_output_bias"],
        )
        
    def _configure_torch_modules(self, func, kwargs):
        
        kwargs['state_size'] = self.hparams['latent_dim']
        
        self._configure_encoder()
        self.func = func(**utils.function_kwargs(func, kwargs))
        self._configure_decoder()
        
    def _configure_optimizers_schedulers(self):
        
        """Assumes use of a pre-train step with an independent optimizer, scheduler (i.e., two total, each)"""
        
        pretrain_optimizer = self.hparams['pretrain_optimizer']
        train_optimizer = self.hparams['train_optimizer']
        pretrain_scheduler = self.hparams['pretrain_scheduler']
        train_scheduler = self.hparams['train_scheduler']
        
        VAE_params = list(self.Encoder.parameters()) + list(self.Decoder.parameters())
        self._optimizers = [
            pretrain_optimizer(VAE_params, lr=self.hparams['pretrain_lr']),
            train_optimizer(self.parameters(), lr=self.hparams['train_lr']),
        ]
        
        self._schedulers = [
            pretrain_scheduler(
                optimizer=self._optimizers[0], step_size=self.hparams['pretrain_step_size']
            ),
            train_scheduler(
                    optimizer=self._optimizers[1], step_size=self.hparams['train_step_size']
                ),
        ]
        
    def encode(self, X: torch.Tensor) -> torch.Tensor:
        return self.Encoder(X)

    def decode(self, Z: torch.Tensor) -> torch.Tensor:
        return self.Decoder(Z)

    def forward(self, X0, t, logqp=False, **kwargs):
        """Forward step: (0) integrate in latent space"""
        Z_hat = self.integrate(
            Z0=self.encode(X0),
            t=t,
            dt=self.hparams["dt"],
            logqp=logqp,
            **kwargs,
        )
        return self.decode(Z_hat)

    def pretrain_step(self, batch, batch_idx, stage=None):
        """Encode only X0->Z0. Return X0_hat"""
        
        pretrain_optim = self.optimizers()[0]
        pretrain_optim.zero_grad()
        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.decode(self.encode(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        
        self.manual_backward(recon_loss)
        pretrain_optim.step()

    def __repr__(self):
        return "scDiffEq MixIn: VAEMixIn"




__module_name__ = "__init__.py"
__version__ = "0.0.46rc3"
__doc__ = """Top-level __init__ for the scdiffeq package."""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# -- import model API: -------------------------------------------------------------------
from .core._scdiffeq import scDiffEq


# -- import sub-packages: ----------------------------------------------------------------
from . import core
from . import io
from . import plotting as pl
from . import tools as tl




import anndata

def idx_to_int_str(adata: anndata.AnnData) -> anndata.AnnData:
    """
    """
    adata.obs = adata.obs.reset_index()
    adata.obs.index = adata.obs.index.astype(str)
    return adata



__module_name__ = "__init__.py"
__version__ = "0.0.45"
__doc__ = """I/O __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# import functions accessed as sdq.io.<func>: --------------------------------------------
from ._read_h5ad import read_h5ad
from ._pickle_io import read_pickle, write_pickle



__module_name__ = "__init__.py"
__doc__ = """TODO"""
__author__ = ", ".join(["Michael E. Vinyard", "Anders Rasmussen", "Ruitong Li"])
__email__ = ", ".join(
    [
        "mvinyard@broadinstitute.org",
        "arasmuss@broadinstitute.org",
        "ruitong@broadinstitute.org",
    ]
)


# -- import: -----------------------------------------------------------------------------
from ._function_kwargs import extract_func_kwargs, function_kwargs
from ._logging_learnable_hparams import LoggingLearnableHParams
from ._autoparse_base_class import AutoParseBase
from ._sum_normalize import sum_normalize

from ._scdiffeq_logger import scDiffEqLogger

from ._default_neural_sde import default_NeuralSDE

from ._normalize_time_to_range import normalize_time

from ._not_none_type import not_none
from ._function_fetch import FunctionFetch, fetch_optimizer, fetch_lr_scheduler


from ._anndata_inspector import AnnDataInspector
from ._info_message import InfoMessage
from ._abc_parse import ABCParse
from ._fast_graph import FastGraph
from ._fetch_format import fetch_format

from ._idx_to_int_str import idx_to_int_str

from ._logs import Logs, PretrainLogs, TrainLogs

from ._filter_df import filter_df

from ._display_tracked_loss import display_tracked_loss

from ._knn_graph_query import kNNGraphQuery


import setuptools
import re
import os
import sys

setuptools.setup(
    name="scdiffeq",
    version="0.0.46rc3",
    python_requires=">3.9.0",
    author="Michael E. Vinyard - Harvard University - Massachussetts General Hospital - Broad Institute of MIT and Harvard",
    author_email="mvinyard@broadinstitute.org",
    url="https://github.com/mvinyard/sc-neural-diffeqs",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    description="scDiffEq: modelling single-cell dynamics using neural differential equations.",
    packages=setuptools.find_packages(),
    install_requires=[
        "anndata>=0.9.1",
        "torch>=2.0.0",
        "pykeops>=2.1",
        "geomloss>=0.2.5",
        "licorice_font>=0.0.3",
        "torchsde>=0.2.5",
        "scikit-learn>=1.0.2",
        "umap-learn>=0.5.3",
        "lightning>=2.0.1",
        "neural-diffeqs==0.3.1rc0",
        "torch-nets>=0.0.4",
        "torch-adata>=0.0.23",
        "autodevice>=0.0.2",
        "brownian-diffuser>=0.0.2",
        "vinplots>=0.0.75",
        "annoyance==0.0.18",
        "ABCParse==0.0.3",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="MIT",
)



from datetime import datetime
import os, glob

from ._autoparse_base_class import AutoParseBase

class scDiffEqLogger(AutoParseBase):
    """
    While Lightning uses automatic logging, we need something one step removed from this to take full
    advantage of their setup within the constraints of our model.
    """

    def __init__(self, model_name="scDiffEq_model", working_dir=os.getcwd()):

        self.__parse__(locals(), public=[None])
        self.creation_count = 0
        self._glob_init = glob.glob(self.default_model_outdir + "/version*")

    @property
    def model_name(self):
        return self._model_name

    @property
    def wd(self):
        return self._working_dir

    @property
    def default_model_outdir(self):
        return os.path.join(self.wd, self.model_name)

    def configure_model_outdir(self):
        if not os.path.exists(self.default_model_outdir):
            os.mkdir(self.default_model_outdir)
            line = "".join(
                [
                    "\n ------------- scDiffEq -------------\n\n\n",
                    f" ---- {datetime.now()} ----\n\n",
                ]
            )

            f = open(self.log_path, mode="w")
            f.write(line)
            f.close()

    @property
    def existing_versions(self):
        return self._glob_init

    @property
    def versioned_model_outdir(self):
        return os.path.join(
            self.default_model_outdir,
            "version_{}".format(len(self.existing_versions)),
        )

    @property
    def log_path(self):
        return os.path.join(self.default_model_outdir, "scDiffEq.log")

    def configure_versioned_model_outdir(self):
        if not os.path.exists(self.versioned_model_outdir):
            if not self.creation_count:
                os.mkdir(self.versioned_model_outdir)
                f = open(self.log_path, mode="a")
                v_path = self.versioned_model_outdir
                v = os.path.basename(v_path)
                line = f"\t{v}\t{datetime.now()}\t{v_path}\n"
                f.write(line)
                f.close()
                self.creation_count += 1

        else:
            print(f"Directory: {self.versioned_model_outdir} already exists!")

    def __call__(self):
        self.configure_model_outdir()
        self.configure_versioned_model_outdir()
